import dill

def unpickle(path):
    with open(path, 'rb') as file:
        return dill.load(file)

def pickle(path, value):
    with open(path, 'wb') as file:
        dill.dump(value, file)

def pickle_string(obj):
    return dill.dumps(obj)




import glob

from funes.pickling import unpickle


def load_cached_results(fn):
    cache_path = fn.__cache_path__
    for file_path in glob.glob(cache_path + '*'):
        res = unpickle(file_path)
        yield res


def load_cached_results_as_pandas(fn, exclude=None, index=None):
    import glob, pandas
    cache_path = fn.__cache_path__
    records = []
    for file_path in glob.glob(cache_path + '*'):
        res = unpickle(file_path)
        inputs = res['input']
        outputs = res['output']
        if not isinstance(outputs, dict):
            outputs = {'output': outputs}
        record = inputs
        record.update(outputs)
        record['timing'] = res['timing']
        records.append(record)
    return pandas.DataFrame.from_records(records, exclude=exclude, index=index)



import opcode
import dis
import weakref
import types

from funes.logging import logging

# this will store hashes and global lists for functions / values,
# without preventing them from being GC'd
code_object_global_names_cache = weakref.WeakKeyDictionary()


STORE_GLOBAL = opcode.opmap['STORE_GLOBAL']
DELETE_GLOBAL = opcode.opmap['DELETE_GLOBAL']
LOAD_GLOBAL = opcode.opmap['LOAD_GLOBAL']
GLOBAL_OPS = (STORE_GLOBAL, DELETE_GLOBAL, LOAD_GLOBAL)


# adapted from https://web.archive.org/web/20140626004012/http://www.picloud.com
def get_code_object_global_names(code_object):
    global code_object_global_names_cache
    out_names = code_object_global_names_cache.get(code_object)
    if out_names:
        return out_names
    try:
        names = code_object.co_names
    except AttributeError:
        out_names = set()
    else:
        out_names = set()
        for instr in dis.get_instructions(code_object):
            op = instr.opcode
            if op in GLOBAL_OPS:
                out_names.add(names[instr.arg])
        # see if nested function have any global refs
        if code_object.co_consts:
            for const in code_object.co_consts:
                if type(const) is types.CodeType:
                    out_names |= set(get_code_object_global_names(const))
    out_names = sorted(out_names)
    code_object_global_names_cache[code_object] = out_names
    return out_names


# this generates the digests of all the globals that a function
# depends on
def hash_code_object_dependencies(fn, hasher, module_scope):
    from funes.hashing import hash_digest
    code_object = fn.__code__
    global_dict = fn.__globals__
    if module_scope and fn.__module__ != module_scope:
        if logging:
            print (f"skipping dependencies out-of-scope function {fn.__name__}")
        return
    for var in get_code_object_global_names(code_object):
        if var in sorted(global_dict.keys()):
            val = global_dict[var]
            digest = hash_digest(val, module_scope)
            if logging:
                print(f'{fn.__name__} depends on {var} which has digest {digest}')
            hasher.update(digest)


# we don't want our hashing of functions to depend on unstable properties of the
# co, like the filename, first line, source map, etc.
def get_stable_code_object_fields(co):
    return (co.co_argcount,
            co.co_nlocals,
            co.co_flags & ~1,   # null out the 'optimized' flag
            co.co_stacksize,
            co.co_names,
            co.co_varnames,
            co.co_code,
            co.co_consts)



import hashlib
import weakref

from funes.pickling import pickle_string
from funes.logging import logging
from funes.co_tools import get_stable_code_object_fields, hash_code_object_dependencies


stable_hash_cache = weakref.WeakKeyDictionary()


# this hashes a value, using dill. if that value is a function, we special case
# the hashing to 1) be stable 2) include the hashes of those functions it depends on
def hash_digest(x, module_scope=None, hasher=None, cache_hash=True):
    global stable_hash_cache
    if cache_hash and x in stable_hash_cache:
        return stable_hash_cache.get(x)
    is_func = hasattr(x, '__code__')
    if hasher is None:
        hasher = hashlib.new('md5')
    if is_func:
        dump = pickle_string(get_stable_code_object_fields(x.__code__))
    else:
        dump = pickle_string(x)
    hasher.update(dump)
    if logging:
        print(f"base hash for {x}: {hasher.hexdigest()}")
    if is_func:
        hash_code_object_dependencies(x, hasher, module_scope or x.__module__)
    digest = hasher.digest()
    if cache_hash:
    	try:
        	stable_hash_cache[x] = digest
    	except TypeError:
        	pass
    return digest


def hash_hexdigest(x, module_scope=None, cache_hash=True):
    hasher = hashlib.new('md5')
    hash_digest(x, module_scope=module_scope, hasher=hasher, cache_hash=cache_hash)
    return hasher.hexdigest()



import os.path
import time
import functools
import sys

from funes.pickling import pickle, unpickle
from funes.hashing import hash_hexdigest
from funes.logging import logging


# this simulates the way that python will resolve positional and named arguments, yielding
# a single ordered dict that contains the names of arguments and their values
def normalize_args(fn_name, args, kwargs, arg_names, defaults):
    result = {}
    for key in kwargs.keys():
        if key not in arg_names:
            if key == 'global_seed':
                result['global_seed'] = kwargs[key]
            else:
                raise RuntimeError(f"{fn_name}: unknown key {key} provided")
    if len(args) > len(arg_names):
        raise RuntimeError(f"{fn_name}: excess arguments provided {len(args)} > {len(arg_names)}")
    for i, name in enumerate(arg_names):
        if i < len(args):
            result[name] = args[i]
        elif name in kwargs:
            result[name] = kwargs[name]
        elif name in defaults:
            result[name] = defaults[name]
        else:
            raise RuntimeError(f"{fn_name}: value for argument {i} ('{name}') not specified")
    return result


def apply_global_seed(seed):
    import random
    random.seed(seed)
    if 'numpy' in sys.modules:
        import numpy
        numpy.random.seed(seed)
    if 'torch' in sys.modules:
        import torch
        torch.manual_seed


# this is the decorator that turns a function into a disk-memoizing version
def memorious(fn):

    fn.__cache_path__ = 'cache/' + fn.__name__ + '/' + hash_hexdigest(fn) + '/'
    fn.__arg_names__ = fn.__code__.co_varnames[:fn.__code__.co_argcount]
    defaults = fn.__defaults__ or []
    fn.__kwdefaults__ = dict(zip(fn.__arg_names__[-len(defaults):], defaults))

    if not os.path.exists(fn.__cache_path__):
        os.makedirs(fn.__cache_path__)

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        input_dict = normalize_args(fn.__name__, args, kwargs, fn.__arg_names__, fn.__kwdefaults__)
        if logging:
            print(f"Input: {input_dict}")
        output_path = fn.__cache_path__ + hash_hexdigest(input_dict, cache_hash=False)
        if os.path.exists(output_path):
            res = unpickle(output_path)
            return res['output']
        start = time.time()
        #
        if 'global_seed' in kwargs:
            apply_global_seed(kwargs['global_seed'])
            del kwargs['global_seed']
        output = fn(*args, **kwargs)
        end = time.time()
        output_dict = {'input': input_dict, 'output': output, 'timing': end - start}
        if logging:
            print(f"Output: {output_dict}")
        pickle(output_path, output_dict)
        return output

    return wrapper


from setuptools import setup

import pathlib

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="funes",
    packages=["funes"],
    install_requires=["dill"],
    version="0.0.1",
    license="MIT",
    description="Decorator to memorize function computations permanently to disk.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/taliesinb/funes",
    download_url="https://github.com/taliesinb/funes/archive/v0.0.1.tar.gz",
    author="Taliesin Beynon",
    author_email="contact@taliesin.ai",
    keywords=["persistent", "cache", "hash", "stable", "archive", "memoize"],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ]
)